import torch


@torch.jit.script
def _one_hot(x: torch.Tensor, num_classes: int = -1, dtype: torch.dtype = torch.float32):
   
    if x.dim() != 2:
        raise ValueError("Input tensor x must be 2D")
    
    if num_classes < 0:
        num_classes = x.max() + 1
    res = torch.zeros(x.shape[0], x.shape[1], num_classes, device=x.device, dtype=dtype)
    tmp = torch.meshgrid(
        torch.arange(x.shape[0], device=x.device),
        torch.arange(x.shape[1], device=x.device),
        indexing="ij",
    )
    index = (tmp[0], tmp[1], x)
    values = torch.ones(x.shape[0], x.shape[1], device=x.device, dtype=dtype)
    res.index_put_(index, values)
    
    return res


def one_hot(x: torch.Tensor, num_classes: int = -1, dtype: torch.dtype = torch.float32):
    """A fast one-hot encoding function faster than the PyTorch one working with torch.int32 and returning a float Tensor.
    Works only for 2D tensors.
    
    Args:
        x (torch.Tensor): Input tensor to be one-hot encoded.
        num_classes (int, optional): Number of classes. If -1, the number of classes is inferred from the input tensor. Defaults to -1.
        dtype (torch.dtype, optional): Data type of the output tensor. Defaults to torch.float32.
        
    Returns:
        torch.Tensor: One-hot encoded tensor.
    """
    return _one_hot(x, num_classes, dtype)


from typing import Dict, Callable

import torch
from torch.nn.functional import one_hot as one_hot_torch

from adabmDCA.functional import one_hot


@torch.jit.script
def _gibbs_sweep(
    chains: torch.Tensor,
    residue_idxs: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float,
) -> torch.Tensor:
    """Performs a Gibbs sweep over the chains.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        residue_idxs (torch.Tensor): List of residue indices in random order.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        beta (float): Inverse temperature.

    Returns:
        torch.Tensor: Updated chains.
    """
    N, L, q = chains.shape
    for i in residue_idxs:
        # Select the couplings attached to the residue we are considering (i) and flatten along the other residues ({j})
        couplings_residue = params["coupling_matrix"][i].view(q, L * q)
        # Update the chains
        logit_residue = beta * (params["bias"][i].unsqueeze(0) + chains.reshape(N, L * q) @ couplings_residue.T) # (N, q)
        chains[:, i, :] = one_hot(torch.multinomial(torch.softmax(logit_residue, -1), 1), num_classes=q).to(logit_residue.dtype).squeeze(1)
        
    return chains


def gibbs_sampling(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    nsweeps: int,
    beta: float = 1.0,
) -> torch.Tensor:
    """Gibbs sampling.
    
    Args:
        chains (torch.Tensor): Initial chains.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        nsweeps (int): Number of sweeps.
        beta (float, optional): Inverse temperature. Defaults to 1.0.
    
    Returns:
        torch.Tensor: Updated chains.
    """
    L = params["bias"].shape[0]
    
    for t in torch.arange(nsweeps):
        # Random permutation of the residues
        residue_idxs = torch.randperm(L)
        chains = _gibbs_sweep(chains, residue_idxs, params, beta)
        
    return chains


def _get_deltaE(
        idx: int,
        chain: torch.Tensor,
        residue_old: torch.Tensor,
        residue_new: torch.Tensor,
        params: Dict[str, torch.Tensor],
        L: int,
        q: int,
    ) -> float:
    
        coupling_residue = chain.view(-1, L * q) @ params["coupling_matrix"][:, :, idx, :].view(L * q, q) # (N, q)
        E_old = - residue_old @ params["bias"][idx] - torch.vmap(torch.dot, in_dims=(0, 0))(coupling_residue, residue_old)
        E_new = - residue_new @ params["bias"][idx] - torch.vmap(torch.dot, in_dims=(0, 0))(coupling_residue, residue_new)
        
        return E_new - E_old
    

def _metropolis_sweep(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float,
) -> torch.Tensor:
    """Performs a Metropolis sweep over the chains.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        beta (float): Inverse temperature.

    Returns:
        torch.Tensor: Updated chains.
    """
    
    N, L, q = chains.shape
    residue_idxs = torch.randperm(L)
    for i in residue_idxs:
        res_old = chains[:, i, :]
        res_new = one_hot_torch(torch.randint(0, q, (N,), device=chains.device), num_classes=q).type(chains.dtype)
        delta_E = _get_deltaE(i, chains, res_old, res_new, params, L, q)
        accept_prob = torch.exp(- beta * delta_E).unsqueeze(-1)
        chains[:, i, :] = torch.where(accept_prob > torch.rand((N, 1), device=chains.device, dtype=chains.dtype), res_new, res_old)

    return chains
    

def metropolis(
    chains: torch.Tensor,
    params: Dict[str, torch.Tensor],
    nsweeps: int,
    beta: float = 1.0,
) -> torch.Tensor:
    """Metropolis sampling.

    Args:
        chains (torch.Tensor): One-hot encoded sequences.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        nsweeps (int): Number of sweeps to be performed.
        beta (float, optional): Inverse temperature. Defaults to 1.0.

    Returns:
        torch.Tensor: Updated chains.
    """

    for _ in range(nsweeps):
        chains = _metropolis_sweep(chains, params, beta)

    return chains


def get_sampler(sampling_method: str) -> Callable:
    """Returns the sampling function corresponding to the chosen method.

    Args:
        sampling_method (str): String indicating the sampling method. Choose between 'metropolis' and 'gibbs'.

    Raises:
        KeyError: Unknown sampling method.

    Returns:
        Callable: Sampling function.
    """
    if sampling_method == "gibbs":
        return gibbs_sampling
    elif sampling_method == "metropolis":
        return metropolis
    else:
        raise KeyError("Unknown sampling method. Choose between 'metropolis' and 'gibbs'.")



import argparse
from pathlib import Path
import numpy as np

import torch

from adabmDCA.fasta import get_tokens
from adabmDCA.io import load_params
from adabmDCA.utils import set_zerosum_gauge, get_device, get_dtype
from adabmDCA.parser import add_args_contacts

# import command-line input arguments
def create_parser():
    # Important arguments
    parser = argparse.ArgumentParser(description='Computes the Frobenius norm matrix extracted from a DCA model.')
    parser = add_args_contacts(parser)
    
    return parser


def main():
    print("\n" + "".join(["*"] * 10) + f" Computing the Frobenius norm " + "".join(["*"] * 10) + "\n")
    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    
    # Check if the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")
    
    # Import parameters
    tokens = get_tokens(args.alphabet)
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(args.path_params, tokens=tokens, device=device, dtype=dtype)
    L, q = params["bias"].shape
    
    # Zero-sum gauge
    params = set_zerosum_gauge(params)
    
    # Get index of the gap symbol
    gap_idx = tokens.index("-")
    
    cm_reduced = params["coupling_matrix"]
    # Take all the entries of the coupling matrix except where the gap is involved
    mask = torch.arange(q) != gap_idx
    cm_reduced = cm_reduced[:, mask, :, :][:, :, :, mask]
    
    # Compute the Frobenius norm
    print("Computing the Frobenius norm...")
    F = torch.sqrt(torch.square(cm_reduced).sum([1, 3]))
    # Set to zero the diagonal
    F = F - torch.diag(F.diag())
    # Compute the average-product corrected Frobenius norm
    Fapc = F - torch.outer(F.sum(1), F.sum(0)) / F.sum()
    
    # Save the results
    print("Saving results...")
    if args.label is not None:
        fname_out = args.output / Path(f"{args.label}_frobenius.txt")
    else:
        fname_out = args.output / Path(f"frobenius.txt")
      
    with open(fname_out, "w") as f:  
        for i in range(Fapc.shape[0]):
            for j in range(Fapc.shape[1]):
                f.write(f"{i},{j},{Fapc[i, j]}\n")

    print(f"Process completed. Results saved in {fname_out}")
    
if __name__ == "__main__":
    main()     
    


from typing import Tuple, Dict, Callable
from tqdm.autonotebook import tqdm

import torch


@torch.jit.script
def get_mean_seqid(
    a1: torch.Tensor,
    a2: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Computes the mean and the standard deviation of the mean sequence identity between two sets of one-hot encoded sequences.

    Args:
        a1 (torch.Tensor): Sequence dataset 1.
        a2 (torch.Tensor): Sequence dataset 2.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Mean sequence identity and standard deviation of the mean.
    """
    
    a1 = a1.view(a1.shape[0], -1)
    a2 = a2.view(a2.shape[0], -1)
    overlaps = (a1 * a2).sum(1)
    mean_overlap = overlaps.mean()
    std_overlap = overlaps.std() / torch.sqrt(overlaps.shape[0])
    
    return mean_overlap, std_overlap


def compute_seqID(a1: torch.Tensor, single_seq: torch.Tensor):
    """
    Computes the Hamming distance 
    between a set of one-hot encoded sequences and a single one-hot encoded sequence.

    Args:
        a1 (torch.Tensor): Sequence dataset, shape (N, L, C), where N is the number of sequences,
                           L is the length, and C is the number of categories (one-hot size).
        single_seq (torch.Tensor): Single one-hot encoded sequence, shape (L, C).

    Returns:
        torch.Tensor: Hamming distances for each sequence in the dataset.
    """
    # print(a1.shape, single_seq.shape)
    a1 = a1.view(a1.shape[0], -1)
    single_seq = single_seq.view(1, -1)
    # print(a1.shape, single_seq.shape)
    seqID = (a1 * single_seq).sum(1) 

    return seqID


def compute_mixing_time(
    sampler: Callable,
    data: torch.Tensor,
    params: Dict[str, torch.Tensor],
    n_max_sweeps: int,
    beta: float,
) -> Dict[str, list]:
    """Computes the mixing time using the t and t/2 method. The sampling will halt when the mixing time is reached or
    the limit of `n_max_sweeps` sweeps is reached.

    Args:
        sampler (Callable): Sampling function.
        data (torch.Tensor): Initial data.
        params (Dict[str, torch.Tensor]): Parameters for the sampling.
        n_max_sweeps (int): Maximum number of sweeps.
        beta (float): Inverse temperature for the sampling.

    Returns:
        Dict[str, list]: Results of the mixing time analysis.
    """

    torch.manual_seed(0)
    
    L, _ = params["bias"].shape
    # Initialize chains at random
    sample_t = data
    # Copy sample_t to a new variable sample_t_half
    sample_t_half = sample_t.clone()

    # Initialize variables
    results = {
        "seqid_t": [],
        "std_seqid_t": [],
        "seqid_t_t_half": [],
        "std_seqid_t_t_half": [],
        "t_half": [],
    }

    # Loop through sweeps
    pbar = tqdm(
        total=n_max_sweeps,
        colour="red",
        dynamic_ncols=True,
        leave=False,
        ascii="-#",
    )
    pbar.set_description("Iterating until the mixing time is reached")
        
    for i in range(1, n_max_sweeps + 1):
        pbar.update(1)
        # Set the seed to i
        torch.manual_seed(i)
        # Perform a sweep on sample_t
        sample_t = sampler(chains=sample_t, params=params, nsweeps=1, beta=beta)

        if i % 2 == 0:
            # Set the seed to i/2
            torch.manual_seed(i // 2)
            # Perform a sweep on sample_t_half
            sample_t_half = sampler(chains=sample_t_half, params=params, nsweeps=1, beta=beta)

            # Calculate the average distance between sample_t and itself shuffled
            perm = torch.randperm(len(sample_t))
            seqid_t, std_seqid_t = get_mean_seqid(sample_t, sample_t[perm])
            seqid_t, std_seqid_t = seqid_t / L, std_seqid_t / L

            # Calculate the average distance between sample_t and sample_t_half
            seqid_t_t_half, std_seqid_t_t_half = get_mean_seqid(sample_t, sample_t_half)
            seqid_t_t_half, std_seqid_t_t_half = seqid_t_t_half / L, std_seqid_t_t_half / L

            # Store the results
            results["seqid_t"].append(seqid_t.item())
            results["std_seqid_t"].append(std_seqid_t.item())
            results["seqid_t_t_half"].append(seqid_t_t_half.item())
            results["std_seqid_t_t_half"].append(std_seqid_t_t_half.item())
            results["t_half"].append(i // 2)

            # Check if they have crossed
            if torch.abs(seqid_t - seqid_t_t_half) / torch.sqrt(std_seqid_t**2 + std_seqid_t_t_half**2) < 0.1:
                break

        if i == n_max_sweeps:
            print(f"Mixing time not reached within {n_max_sweeps // 2} sweeps.")
            
    pbar.close()

    return results


# Converts the old DCA parameters format in the new human-readable format that does not require to specify the alphabet
# (i, j, a, b, value) -> (i, j, aa(a), aa(b), value)

import argparse
from pathlib import Path
import numpy as np

from adabmDCA.fasta import get_tokens
from adabmDCA.io import load_params_oldformat, save_params


# import command-line input arguments
def create_parser():
    parser = argparse.ArgumentParser(description="Converts the old DCA parameters format in the new human-readable format that does not require to specify the alphabet (i, j, a, b, value) -> (i, j, aa(a), aa(b), value)")
    parser.add_argument("-p", "--path_params",  type=Path,   required=True,      help="Path to the file containing the parameters of DCA model to sample from.")
    parser.add_argument("--alphabet",           type=str,    default="protein",  help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    return parser


def main():       
    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    # Import parameters in the old format
    params = load_params_oldformat(args.path_params, device="cpu") # This function will be modified
    tokens = get_tokens(args.alphabet)
    
    # Save parameters in the new format
    fname_out = args.path_params.parent / Path(args.path_params.stem + "_newformat.dat")
    L, q, *_ = params["coupling_matrix"].shape
    mask1 = np.zeros(shape=(L, q, L, q))
    mask1[np.nonzero(params["coupling_matrix"])] = 1
    mask2 = np.ones(shape=(L, q, L, q))
    idx1_rm, idx2_rm = np.tril_indices(L, k=0)
    mask2[idx1_rm, :, idx2_rm, :] = np.zeros(shape=(q, q))
    save_params(fname_out, params, np.logical_and(mask1, mask2), tokens=tokens)
    print(f"Completed. Output saved in {fname_out}")
    
    
if __name__ == '__main__':
    main()


from typing import Dict, Tuple
import pandas as pd
import numpy as np

import torch

from adabmDCA.fasta import (
    write_fasta,
    encode_sequence,
    import_from_fasta,
    validate_alphabet,
    get_tokens,
)


def load_chains(
    fname: str,
    tokens: str,
    load_weights: bool = False
) -> np.ndarray | Tuple[np.ndarray, np.ndarray]:
    """Loads the sequences from a fasta file and returns the numeric-encoded version.
    If the sequences are weighted, the log-weights are also returned. If the sequences are not weighted, the log-weights are set to 0.
    
    Args:
        fname (str): Path to the file containing the sequences.
        tokens (str): "protein", "dna", "rna" or another string with the alphabet to be used.
        load_weights (bool, optional): If True, the log-weights are loaded and returned. Defaults to False.
    
    Return:
        np.ndarray | Tuple[np.ndarray, np.ndarray]: Numeric-encoded sequences and log-weights if load_weights is True.
    """
    def parse_header(header: str):
        h = header.split("|")
        if len(h) == 2:
            log_weigth = float(h[1].split("=")[1])
            return log_weigth
        else:
            return 0.
    
    headers, sequences = import_from_fasta(fasta_name=fname)
    validate_alphabet(sequences, tokens=tokens)
    encoded_sequences = encode_sequence(sequences, tokens=tokens)
    
    if load_weights:
        log_weights = np.vectorize(parse_header)(headers)
        return encoded_sequences, log_weights
    else:
        return encoded_sequences
    

def save_chains(
    fname: str,
    chains: torch.Tensor,
    tokens: str,
    log_weights: torch.Tensor = None
) -> None:
    """Saves the chains in a fasta file.

    Args:
        fname (str): Path to the file where to save the chains.
        chains (torch.Tensor): Chains.
        tokens (str): "protein", "dna", "rna" or another string with the alphabet to be used.
        log_weights (torch.Tensor, optional): Log-weights of the chains. Defaults to None.
    """
    
    # Check if chains is a 3D tensor
    if chains.ndim != 2:
        raise ValueError("chains must be a 2D tensor")
    
    chains = chains.cpu().numpy()
    if log_weights is not None:
        log_weigth = log_weights.cpu().numpy()
        headers = [f"chain_{i}|log_weight={log_weigth[i]}" for i in range(len(chains))]
    else:
        headers = [f"chain_{i}" for i in range(len(chains))]
    write_fasta(
        fname=fname,
        headers=headers,
        sequences=chains,
        numeric_input=True,
        remove_gaps=False,
        tokens=tokens,
    )
    
    
def load_params(
    fname: str,
    tokens: str,
    device: torch.device,
    dtype: torch.dtype = torch.float32,
) -> Dict[str, torch.Tensor]:
    """Import the parameters of the model from a file.

    Args:
        fname (str): Path of the file that stores the parameters.
        tokens (str): "protein", "dna", "rna" or another string with the alphabet to be used.
        device (torch.device): Device where to store the parameters.
        dtype (torch.dtype): Data type of the parameters. Defaults to torch.float32.

    Returns:
        Dict[str, torch.Tensor]: Parameters of the model.
    """
    
    param_labels = pd.read_csv(fname, sep=" ", usecols=[0,]).to_numpy()
    skiprows = (param_labels == "J").sum() + 1
    skipfooter = len(param_labels) - skiprows + 1

    df_J = pd.read_csv(
        fname,
        sep=" ",
        names=["param", "idx0", "idx1", "idx2", "idx3", "val"],
        skipfooter=skipfooter,
        engine="python"
    ).astype({"idx0" : int, "idx1" : int, "idx2" : str, "idx3" : str, "val" : float})
    
    df_h = pd.read_csv(
        fname,
        sep=" ",
        names=["param", "idx0", "idx1", "val"],
        skiprows=skiprows
    ).astype({"idx0" : int, "idx1" : str, "val" : float})
    
    # Convert from amino acid format to numeric format
    tokens = get_tokens(tokens)
    validate_alphabet(df_h["idx1"].to_numpy(), tokens=tokens)
    df_J["idx2"] = encode_sequence(df_J["idx2"].to_numpy(), tokens=tokens)
    df_J["idx3"] = encode_sequence(df_J["idx3"].to_numpy(), tokens=tokens)
    df_h["idx1"] = encode_sequence(df_h["idx1"].to_numpy(), tokens=tokens)
    

    h_idx = df_h.loc[:, ["idx0", "idx1"]].to_numpy()
    L, q = h_idx.max(0) + 1
    h_val = df_h.loc[:, "val"].to_numpy()

    h = np.zeros(shape=(L * q,))
    h_idx_flat = h_idx @ np.array([q, 1])
    h[h_idx_flat] = h_val
    h = h.reshape(L, q)
    
    J_idx = df_J.loc[:, ["idx0", "idx1", "idx2", "idx3"]].to_numpy()
    J_val = df_J.loc[:, "val"].to_numpy()

    J = np.zeros(shape=(L**2 * q**2,))
    J_idx_flat = J_idx @ np.array([L * q**2, q**2, q, 1])
    J[J_idx_flat] = J_val
    
    # Only the upper-triangular part of J is filled
    J = J.reshape(L, L, q, q).transpose(0, 2, 1, 3).reshape(L * q, L * q)
    J = (J + J.T).reshape(L, q, L, q)

    return {
        "bias" : torch.tensor(h, dtype=dtype, device=device),
        "coupling_matrix" : torch.tensor(J, dtype=dtype, device=device),
        }


def save_params(
    fname: str,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
    tokens: str,
) -> None:
    """Saves the parameters of the model in a file.

    Args:
        fname (str): Path to the file where to save the parameters.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        mask (torch.Tensor): Mask of the coupling matrix that determines which are the non-zero entries.
        tokens (str): "protein", "dna", "rna" or another string with the alphabet to be used.
    """
    tokens = get_tokens(tokens)
    mask = mask.cpu().numpy()
    params = {k : v.cpu().numpy() for k, v in params.items()}
    
    L, q, *_ = mask.shape
    idx0 = np.arange(L * q).reshape(L * q) // q
    idx1 = np.arange(L * q).reshape(L * q) % q
    idx1_aa = np.vectorize(lambda n, tokens : tokens[n], excluded=["tokens"])(idx1, tokens).astype(str)
    df_h = pd.DataFrame(
        {
            "param" : np.full(L * q, "h"),
            "idx0" : idx0,
            "idx1" : idx1_aa,
            "idx2" : params["bias"].flatten(),
        }
    )
    
    
    maskt = mask.transpose(0, 2, 1, 3) # Transpose mask and coupling matrix from (L, q, L, q) to (L, L, q, q)
    Jt = params["coupling_matrix"].transpose(0, 2, 1, 3)
    idx0, idx1, idx2, idx3 = maskt.nonzero()
    idx2_aa = np.vectorize(lambda n, tokens : tokens[n], excluded=["tokens"])(idx2, tokens).astype(str)
    idx3_aa = np.vectorize(lambda n, tokens : tokens[n], excluded=["tokens"])(idx3, tokens).astype(str)
    J_val = Jt[idx0, idx1, idx2, idx3]
    df_J = pd.DataFrame(
        {
            "param" : np.full(len(J_val), "J").tolist(),
            "idx0" : idx0,
            "idx1" : idx1,
            "idx2" : idx2_aa,
            "idx3" : idx3_aa,
            "val" : J_val,
        }
    )
    df_J.to_csv(fname, sep=" ", header=False, index=False)
    df_h.to_csv(fname, sep=" ", header=False, index=False, mode="a")
    
    
def load_params_oldformat(
    fname: str,
    device: torch.device,
    dtype: torch.dtype = torch.float32,
) -> Dict[str, torch.Tensor]:
    """Import the parameters of the model from a file. Assumes the old DCA format.

    Args:
        fname (str): Path of the file that stores the parameters.
        device (torch.device): Device where to store the parameters.
        dtype (torch.dtype): Data type of the parameters. Defaults to torch.float32.

    Returns:
        Dict[str, torch.Tensor]: Parameters of the model.
    """
    df = pd.read_csv(fname, sep=" ", names=["param", "idx0", "idx1", "idx2", "idx3", "val"])
    df_J = df.loc[df["param"] == "J", ["idx0", "idx1", "idx2", "idx3", "val"]].astype({"idx0" : int, "idx1" : int, "idx2" : int, "idx3" : int, "val" : float})
    df_h = df.loc[df["param"] == "h", ["idx0", "idx1", "idx2"]].astype({"idx0" : int, "idx1" : int, "idx2" : float}).rename(columns={"idx2" : "val"})

    h_idx = df_h.loc[:, ["idx0", "idx1"]].to_numpy()
    L, q = h_idx.max(0) + 1
    h_val = df_h.loc[:, "val"].to_numpy()

    h = np.zeros(shape=(L * q,))
    h_idx_flat = h_idx @ np.array([q, 1])
    h[h_idx_flat] = h_val
    h = h.reshape(L, q)
    
    J_idx = df_J.loc[:, ["idx0", "idx1", "idx2", "idx3"]].to_numpy()
    J_val = df_J.loc[:, "val"].to_numpy()

    J = np.zeros(shape=(L**2 * q**2,))
    J_idx_flat = J_idx @ np.array([L * q**2, q**2, q, 1])
    J[J_idx_flat] = J_val
    
    # Only the upper-triangular part of J is filled
    J = J.reshape(L, L, q, q).transpose(0, 2, 1, 3).reshape(L * q, L * q)
    J = (J + J.T).reshape(L, q, L, q)

    return {
        "bias" : torch.tensor(h, dtype=dtype, device=device),
        "coupling_matrix" : torch.tensor(J, dtype=dtype, device=device),
        }


def save_params_oldformat(
    fname: str,
    params: Dict[str, torch.Tensor],
    mask: torch.Tensor,
) -> None:
    """Saves the parameters of the model in a file. Assumes the old DCA format.

    Args:
        fname (str): Path to the file where to save the parameters.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        mask (torch.Tensor): Mask of the coupling matrix that determines which are the non-zero entries.
    """
    mask = mask.cpu().numpy()
    params = {k : v.cpu().numpy() for k, v in params.items()}
    
    L, q, *_ = mask.shape
    idx0 = np.arange(L * q).reshape(L * q) // q
    idx1 = np.arange(L * q).reshape(L * q) % q
    df_h = pd.DataFrame.from_dict({"param" : np.full(L * q, "h"), "idx0" : idx0, "idx1" : idx1, "idx2" : params["bias"].flatten()}, orient="columns")
    
    maskt = mask.transpose(0, 2, 1, 3) # Transpose mask and coupling matrix from (L, q, L, q) to (L, L, q, q)
    Jt = params["coupling_matrix"].transpose(0, 2, 1, 3)
    idx0, idx1, idx2, idx3 = maskt.nonzero()
    J_val = Jt[idx0, idx1, idx2, idx3]
    df_J = pd.DataFrame.from_dict({"param" : np.full(len(J_val), "J"), "idx0" : idx0, "idx1" : idx1, "idx2" : idx2, "idx3" : idx3, "val" : J_val}, orient="columns")
    df_J.to_csv(fname, sep=" ", header=False, index=False)
    df_h.to_csv(fname, sep=" ", header=False, index=False, mode="a")


import itertools
from typing import Tuple

import torch


@torch.jit.script
def _get_freq_single_point(
    data: torch.Tensor,
    weights: torch.Tensor,
    pseudo_count: float,
) -> torch.Tensor:    
    _, _, q = data.shape
    frequencies = (data * weights).sum(dim=0)
    # Set to zero the negative frequencies. Used for the reintegration.
    torch.clamp_(frequencies, min=0.0)

    return (1. - pseudo_count) * frequencies + (pseudo_count / q)


def get_freq_single_point(
    data: torch.Tensor,
    weights: torch.Tensor | None = None,
    pseudo_count: float = 0.0,
) -> torch.Tensor:
    """Computes the single point frequencies of the input MSA.
    Args:
        data (torch.Tensor): One-hot encoded data array.
        weights (torch.Tensor | None, optional): Weights of the sequences.
        pseudo_count (float, optional): Pseudo count to be added to the frequencies. Defaults to 0.0.
    
    Raises:
        ValueError: If the input data is not a 3D tensor.

    Returns:
        torch.Tensor: Single point frequencies.
    """
    if data.dim() != 3:
        raise ValueError(f"Expected data to be a 3D tensor, but got {data.dim()}D tensor instead")
    M = len(data)
    if weights is not None:
        norm_weights = weights.reshape(M, 1, 1) / weights.sum()
    else:
        norm_weights = torch.ones((M, 1, 1), device=data.device, dtype=data.dtype) / M
    
    return _get_freq_single_point(data, norm_weights, pseudo_count)


@torch.jit.script
def _get_freq_two_points(
    data: torch.Tensor,
    weights: torch.Tensor,
    pseudo_count: float,
) -> torch.Tensor:
    
    M, L, q = data.shape
    data_oh = data.reshape(M, q * L)
    
    fij = (data_oh * weights).T @ data_oh
    # Apply the pseudo count
    fij = (1. - pseudo_count) * fij + (pseudo_count / q**2)
    # Diagonal terms must represent the single point frequencies
    fi = get_freq_single_point(data, weights, pseudo_count).ravel()
    # Apply the pseudo count on the single point frequencies
    fij_diag = (1. - pseudo_count) * fi + (pseudo_count / q)
    # Set the diagonal terms of fij to the single point frequencies
    fij = torch.diagonal_scatter(fij, fij_diag, dim1=0, dim2=1)
    # Set to zero the negative frequencies. Used for the reintegration.
    torch.clamp_(fij, min=0.0)
    
    return fij.reshape(L, q, L, q)


def get_freq_two_points(
    data: torch.Tensor,
    weights: torch.Tensor | None = None,
    pseudo_count: float = 0.0,
) -> torch.Tensor:
    """
    Computes the 2-points statistics of the input MSA.

    Args:
        data (torch.Tensor): One-hot encoded data array.
        weights (torch.Tensor | None, optional): Array of weights to assign to the sequences of shape.
        pseudo_count (float, optional): Pseudo count for the single and two points statistics. Acts as a regularization. Defaults to 0.0.
    
    Raises:
        ValueError: If the input data is not a 3D tensor.

    Returns:
        torch.Tensor: Matrix of two-point frequencies of shape (L, q, L, q).
    """
    if data.dim() != 3:
        raise ValueError(f"Expected data to be a 3D tensor, but got {data.dim()}D tensor instead")
    
    M = len(data)
    if weights is not None:
        norm_weights = weights.reshape(M, 1) / weights.sum()
    else:
        norm_weights = torch.ones((M, 1), device=data.device, dtype=data.dtype) / M
    
    return _get_freq_two_points(data, norm_weights, pseudo_count)


def generate_unique_triplets(
    L: int,
    ntriplets: int,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Generates a set of unique triplets of positions. Used to compute the 3-points statistics.
    
    Args:
        L (int): Length of the sequences.
        ntriplets (int): Number of triplets to be generated.
        device (torch.device, optional): Device to perform computations on. Defaults to "cpu".
    
    Returns:
        torch.Tensor: Tensor of shape (ntriplets, 3) containing the indices of the triplets.
    """    
    # Generate all possible unique triplets
    all_triplets = torch.tensor(list(itertools.combinations(range(L), 3)), device=device)
    # Shuffle the triplets to ensure randomness
    shuffled_triplets = all_triplets[torch.randperm(all_triplets.size(0))]
    # Select the first ntriplets from the shuffled triplets
    selected_triplets = shuffled_triplets[:ntriplets]
    
    return selected_triplets


@torch.jit.script
def _get_C_ijk(
    triplet: torch.Tensor,
    data: torch.Tensor,
    weights: torch.Tensor,
) -> torch.Tensor:
    """
    Computes the third-order correlation tensor for a given triplet of positions.

    Args:
        triplet (torch.Tensor): Tensor of shape (3,) containing the indices of the triplet.
        data (torch.Tensor): Tensor of shape (M, L, q) containing the one-hot encoded data.
        weights (torch.Tensor): Tensor of shape (M, 1) containing the normalized weights of the sequences.

    Returns:
        torch.Tensor: Tensor of shape (q, q, q) containing the third-order correlations.
    """
    # Center the dataset
    data_c = data - data.mean(dim=0, keepdim=True)
    x = data_c[:, triplet[0], :] * weights
    y = data_c[:, triplet[1], :]
    z = data_c[:, triplet[2], :]

    C = torch.einsum("mi, mj, mk -> ijk", x, y, z)

    return C


def get_freq_three_points(
    data: torch.Tensor,
    ntriplets: int,
    weights: torch.Tensor | None = None,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Computes the 3-body statistics of the input MSA.

    Args:
        data (torch.Tensor): Input MSA in one-hot encoding.
        ntriplets (int): Number of triplets to test.
        weights (torch.Tensor | None, optional): Importance weights for the sequences. Defaults to None.
        device (torch.device, optional): Device to perform computations on. Defaults to "cpu".

    Returns:
        torch.Tensor: 3-points connected correlation for ntriplets randomly extracted triplets.
    """
    if data.dim() != 3:
        raise ValueError(f"Expected data to be a 3D tensor, but got {data.dim()}D tensor instead")
    
    M = len(data)
    if weights is not None:
        norm_weights = weights.view(-1, 1) / weights.sum()
    else:
        norm_weights = torch.ones((M, 1), device=data.device, dtype=data.dtype) / M
    
    L = data.shape[1]
    triplets = generate_unique_triplets(L=L, ntriplets=ntriplets, device=device)
    Cijk = []
    for triplet in triplets:
        Cijk.append(_get_C_ijk(triplet, data, norm_weights).flatten())
        
    return torch.stack(Cijk)


def get_covariance_matrix(
    data: torch.Tensor,
    weights: torch.Tensor | None = None,
    pseudo_count: float = 0.0,
) -> torch.Tensor:
    """Computes the weighted covariance matrix of the input multi sequence alignment.

    Args:
        data (torch.Tensor): Input MSA in one-hot variables.
        weights (torch.Tensor | None, optional): Importance weights of the sequences.
        pseudo_count (float, optional): Pseudo count. Defaults to 0.0.

    Returns:
        torch.Tensor: Covariance matrix.
    """
    _, L, q = data.shape
    fi = get_freq_single_point(data, weights, pseudo_count)
    fij = get_freq_two_points(data, weights, pseudo_count)
    cov_matrix = fij - torch.einsum('ij,kl->ijkl', fi, fi)
    
    return cov_matrix.reshape(L * q, L * q)


def _get_slope(x, y):
    n = len(x)
    num = n * (x @ y) - y.sum() * x.sum()
    den = n * (x @ x) - torch.square(x.sum())
    return torch.abs(num / den)


def extract_Cij_from_freq(
    fij: torch.Tensor,
    pij: torch.Tensor,
    fi: torch.Tensor,
    pi: torch.Tensor,
    mask: torch.Tensor | None = None,
) -> Tuple[float, float]:
    """Extracts the lower triangular part of the covariance matrices of the data and chains starting from the frequencies.

    Args:
        fij (torch.Tensor): Two-point frequencies of the data.
        pij (torch.Tensor): Two-point frequencies of the chains.
        fi (torch.Tensor): Single-point frequencies of the data.
        pi (torch.Tensor): Single-point frequencies of the chains.
        mask (torch.Tensor | None, optional): Mask for comparing just a subset of the couplings. Defaults to None.

    Returns:
        Tuple[float, float]: Extracted two-point frequencies of the data and chains.
    """
    L = fi.shape[0]
    
    # Compute the covariance matrices
    cov_data = fij - torch.einsum('ij,kl->ijkl', fi, fi)
    cov_chains = pij - torch.einsum('ij,kl->ijkl', pi, pi)
    
    # Only use a subset of couplings if a mask is provided
    if mask is not None:
        cov_data = torch.where(mask, cov_data, torch.tensor(0.0, device=cov_data.device, dtype=cov_data.dtype))
        cov_chains = torch.where(mask, cov_chains, torch.tensor(0.0, device=cov_chains.device, dtype=cov_chains.dtype))
    
    # Extract only the entries of half the matrix and out of the diagonal blocks
    idx_row, idx_col = torch.tril_indices(L, L, offset=-1)
    fij_extract = cov_data[idx_row, :, idx_col, :].reshape(-1)
    pij_extract = cov_chains[idx_row, :, idx_col, :].reshape(-1)
    
    return fij_extract, pij_extract


def extract_Cij_from_seqs(
    data: torch.Tensor,
    chains: torch.Tensor,
    weights: torch.Tensor | None = None,
    pseudo_count: float = 0.0,
    mask: torch.Tensor | None = None,
) -> Tuple[float, float]:
    """Extracts the lower triangular part of the covariance matrices of the data and chains starting from the sequences.

    Args:
        data (torch.Tensor): Data sequences.
        chains (torch.Tensor): Chain sequences.
        weights (torch.Tensor | None, optional): Weights of the sequences. Defaults to None.
        pseudo_count (float, optional): Pseudo count for the single and two points statistics. Acts as a regularization. Defaults to 0.0.
        mask (torch.Tensor | None, optional): Mask for comparing just a subset of the couplings. Defaults to None.

    Returns:
        Tuple[float, float]: Two-point frequencies of the data and chains.
    """
    fi = get_freq_single_point(data, weights=weights, pseudo_count=pseudo_count)
    pi = get_freq_single_point(chains, weights=None)
    fij = get_freq_two_points(data, weights=weights, pseudo_count=pseudo_count)
    pij = get_freq_two_points(chains, weights=None)
    
    return extract_Cij_from_freq(fij, pij, fi, pi, mask)


def get_correlation_two_points(
    fij: torch.Tensor,
    pij: torch.Tensor,
    fi: torch.Tensor,
    pi: torch.Tensor,
    mask: torch.Tensor | None = None,
) -> Tuple[float, float]:
    """Computes the Pearson coefficient and the slope between the two-point frequencies of data and chains.

    Args:
        fij (torch.Tensor): Two-point frequencies of the data.
        pij (torch.Tensor): Two-point frequencies of the chains.
        fi (torch.Tensor): Single-point frequencies of the data.
        pi (torch.Tensor): Single-point frequencies of the chains.
        mask (torch.Tensor | None, optional): Mask to select the couplings to use for the correlation coefficient. Defaults to None. 

    Returns:
        Tuple[float, float]: Pearson correlation coefficient of the two-sites statistics and slope of the interpolating line.
    """
    
    fij_extract, pij_extract = extract_Cij_from_freq(fij, pij, fi, pi, mask)
    # torch.corrcoef does not support half precision
    pearson = torch.corrcoef(torch.stack([fij_extract.float(), pij_extract.float()]))[0, 1].item()
    slope = _get_slope(fij_extract.float(), pij_extract.float()).item()
    
    return pearson, slope




from typing import List
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import numpy as np

def _plot_scatter_labels(
    ax: plt.Axes,
    data1: np.ndarray,
    pc1: int = 0,
    pc2: int = 1,
    data2: np.ndarray | None = None,
    labels: List[str] | str | None = "Data",
    colors: List[str] | str = "black",
) -> plt.Axes:

    if isinstance(labels, str):
        labels = [labels,]
    if isinstance(colors, str):
        colors = [colors,]
    if labels is None:
        labels = [None, None]
    ax.scatter(data1[:, pc1], data1[:, pc2], color=colors[0], s=50, label=labels[0], zorder=0, alpha=0.3)
    if data2 is not None:
        if len(labels) == 1:
            labels.append("Generated")
        if len(colors) == 1:
            colors.append("red")
        ax.scatter(data2[:, pc1], data2[:, pc2], color=colors[1], label=labels[1], s=20, zorder=2, edgecolor='black', marker='o', alpha=1, linewidth=0.4)
    
    return ax
    
def _plot_hist(
    ax: plt.Axes,
    data1: np.ndarray,
    pc: int,
    data2: np.ndarray | None = None,
    colors: List[str] | str = "black",
    labels: List[str] | str | None = "Data",
    orientation='vertical',
) -> plt.Axes:
    if isinstance(labels, str):
        labels = [labels,]
    if labels is None:
        labels = [None, None]
    if isinstance(colors, str):
        colors = [colors,]
    ax.hist(data1[:, pc], bins=40, color=colors[0], histtype='step', label=labels[0], zorder=0, density=True, orientation=orientation, lw=1)
    if data2 is not None:
        if len(labels) == 1:
            labels.append("Generated")
        if len(colors) == 1:
            colors.append("red")
    if data2 is not None:
        ax.hist(data2[:, pc], bins=40, color=colors[1], histtype='step', label=labels[1], zorder=1, density=True, orientation=orientation, lw=1.5)
    ax.axis('off')
    
    return ax
    
def plot_PCA(
    fig: plt.figure,
    data1: np.ndarray,
    pc1: int = 0,
    pc2: int = 1,
    data2: np.ndarray | None = None,
    labels: List[str] | str = "Data",
    colors: List[str] | str = "black",
    title: str | None = None,
) -> plt.figure:
    """Makes the scatter plot of the components (pc1, pc2) of the input data and shows the histograms of the components.

    Args:
        fig (plt.figure): Figure to plot the data.
        data1 (np.ndarray): Data to plot.
        pc1 (int, optional): First principal direction. Defaults to 0.
        pc2 (int, optional): Second principal direction. Defaults to 1.
        data2 (np.ndarray | None, optional): Data to be superimposed to data1. Defaults to None.
        labels (List[str] | str, optional): Labels to put in the legend. Defaults to "Data".
        colors (List[str] | str, optional): Colors to be used. Defaults to "black".
        title (str | None, optional): Title of the plot. Defaults to None.

    Returns:
        plt.figure: Updated figure.
    """
    
    gs = GridSpec(4, 4)
    ax_scatter = fig.add_subplot(gs[1:4, 0:3])
    ax_hist_x = fig.add_subplot(gs[0, 0:3])
    ax_hist_y = fig.add_subplot(gs[1:4, 3])
        
    ax_scatter = _plot_scatter_labels(
        ax=ax_scatter,
        data1=data1,
        pc1=pc1,
        pc2=pc2,
        data2=data2,
        labels=None,
        colors=colors,
    )
    ax_hist_x = _plot_hist(
        ax=ax_hist_x,
        data1=data1,
        pc=pc1,
        data2=data2,
        colors=colors,
        labels=labels,
        orientation='vertical',
    )
    ax_hist_y = _plot_hist(
        ax=ax_hist_y,
        data1=data1,
        pc=pc2,
        data2=data2,
        colors=colors,
        labels=None,
        orientation='horizontal',
    )
    
    ax_scatter.set_xlabel(f"PC {pc1 + 1}")
    ax_scatter.set_ylabel(f"PC {pc2 + 1}")
    fig.suptitle(title)
    fig.legend(fontsize=12, bbox_to_anchor=(1, 1));
    h, l = ax_scatter.get_legend_handles_labels()
    
    return fig
    
    
def plot_pearson_sampling(
    ax: plt.Axes,
    checkpoints: np.ndarray,
    pearsons: np.ndarray,
    pearson_training: np.ndarray | None = None
):
    
    if pearson_training:
        ax.axhline(y=pearson_training, ls="dashed", color="red", label="Training", lw=1, zorder=0)
        annotation_text = f"Training: {pearson_training:.3f}\nSampling: {pearsons[-1]:.3f}"
    else:
        annotation_text = annotation_text = f"Sampling: {pearsons[-1]:.3f}"
    ax.plot(checkpoints, pearsons, "-o", label="Resampling", lw=0.5, color="royalblue", zorder=1)
    ax.set_xscale("log")
    ax.set_xlabel("Sampling time [sweeps]")
    ax.set_ylabel(r"Pearson $C_{ij}(a,b)$")
    ax.annotate(annotation_text, xy=(0.95, 0.53), xycoords='axes fraction',
                verticalalignment='top', horizontalalignment='right', bbox=dict(facecolor='white', alpha=0.5))
    ax.legend()
    return ax


def plot_autocorrelation(
    ax: plt.Axes,
    checkpoints: np.ndarray,
    autocorr: np.ndarray,
    gen_seqid: np.ndarray,
    data_seqid: np.ndarray
) -> plt.Axes:
    """Plots the time-autocorrelation curve of the sequence identity and the generated and data sequence identities.
    
    Args:
        ax (plt.Axes): Axes to plot the data.
        checkpoints (np.ndarray): Checkpoints of the sampling.
        autocorr (np.ndarray): Time-autocorrelation of the sequence identity.
        gen_seqid (np.ndarray): Sequence identity of the generated data.
        data_seqid (np.ndarray): Sequence identity of the data.
        
    Returns:
        plt.Axes: Updated axes.
    """
    ax.plot(checkpoints, autocorr, "-o", c="royalblue", lw=0.5, label="Time-autocorrelation", zorder=2)
    ax.axhline(y=gen_seqid, color="navy", lw=1, ls="dashed", label="Generated seqID", zorder=1)
    ax.axhline(y=data_seqid, color="red", lw=1, ls="dashed", label="Data seqID", zorder=0)
    ax.set_xlabel(r"$\tau$ [sweeps]")
    ax.set_ylabel(r"$\langle \mathrm{SeqID}(t, t-\tau) \rangle$")
    ax.legend();

    if np.any(checkpoints[autocorr <= gen_seqid]):
        mixing_time = checkpoints[autocorr <= gen_seqid][0]
        print(f"Mixing time: {mixing_time} sweeps")
        annotation_text = r"$\tau_{\mathrm{mix}}$" + f": {mixing_time} sweeps"
        plt.annotate(annotation_text, xy=(0.95, 0.65), xycoords='axes fraction',
                verticalalignment='top', horizontalalignment='right', bbox=dict(facecolor='white', alpha=0.5))
    else:
        print(f"The mixing time could not be computed within {checkpoints[-1]} sweeps")
    
    return ax


def plot_scatter_correlations(
    ax: plt.Axes,
    Cij_data: np.ndarray,
    Cij_gen: np.ndarray,
    Cijk_data: np.ndarray,
    Cijk_gen: np.ndarray,
    pearson_Cij: float,
    pearson_Cijk: float
) -> plt.Axes:
    """Plots the scatter plot of the data and generated Cij and Cijk values.
    
    Args:
        ax (plt.Axes): Axes to plot the data. Must have 2 subplots.
        Cij_data (np.ndarray): Data Cij values.
        Cij_gen (np.ndarray): Generated Cij values.
        Cijk_data (np.ndarray): Data Cijk values.
        Cijk_gen (np.ndarray): Generated Cijk values.
        pearson_Cij (float): Pearson correlation coefficient of Cij.
        pearson_Cijk (float): Pearson correlation coefficient of Cijk.
        
    Returns:
        plt.Axes: Updated axes.
    """
    if len(ax) != 2:
        raise ValueError("The axes must have 2 subplots")
    
    color_line = "#50424F"
    color_scatter = "#FF6275"

    x = np.linspace(Cij_data.min(), Cij_data.max(), 100)
    ax[0].scatter(Cij_data, Cij_gen, alpha=0.5, color=color_scatter)
    ax[0].plot(x, x, ls="dashed", color=color_line)
    ax[0].set_xlabel(r"$C_{ij}$ data")
    ax[0].set_ylabel(r"$C_{ij}$ generated")

    x = np.linspace(Cijk_data.min(), Cijk_data.max(), 100)
    ax[1].scatter(Cijk_data, Cijk_gen, alpha=0.5, color=color_scatter)
    ax[1].plot(x, x, ls="dashed", color=color_line)
    ax[1].set_xlabel(r"$C_{ijk}$ data")
    ax[1].set_ylabel(r"$C_{ijk}$ generated")

    ax[0].annotate(r"$\rho=$" + f"{pearson_Cij:.2f}", xy=(0.05, 0.95), xycoords='axes fraction', fontsize=12,
                    verticalalignment='top', horizontalalignment='left', bbox=dict(facecolor='white', alpha=0.5))

    ax[1].annotate(r"$\rho=$" + f"{pearson_Cijk:.2f}", xy=(0.05, 0.95), xycoords='axes fraction', fontsize=12,
                    verticalalignment='top', horizontalalignment='left', bbox=dict(facecolor='white', alpha=0.5))
    
    return ax


import argparse
from pathlib import Path
import pandas as pd
from tqdm import tqdm

from adabmDCA.fasta import (
    get_tokens,
    write_fasta,
)
from adabmDCA.resampling import compute_mixing_time
from adabmDCA.io import load_params
from adabmDCA.dataset import DatasetDCA
from adabmDCA.utils import init_chains, resample_sequences, get_device, get_dtype
from adabmDCA.sampling import get_sampler
from adabmDCA.functional import one_hot
from adabmDCA.statmech import compute_energy
from adabmDCA.stats import get_freq_single_point, get_freq_two_points, get_correlation_two_points
from adabmDCA.parser import add_args_sample


# import command-line input arguments
def create_parser():
    # Important arguments
    parser = argparse.ArgumentParser(description='Samples from a DCA model.')
    parser = add_args_sample(parser)
    
    return parser


def main():       
    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    # Create output folder
    filename = Path(args.output)
    folder = filename.parent / Path(filename.name)
    # Create the folder where to save the samples
    folder.mkdir(parents=True, exist_ok=True)
    
    print("\n" + "".join(["*"] * 10) + f" Sampling from DCA model " + "".join(["*"] * 10) + "\n")
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    tokens = get_tokens(args.alphabet)
    
    # Check that the data file exists
    if not Path(args.data).exists():
        raise FileNotFoundError(f"Data file {args.data} not found.")
    
    # Check that the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")
        
    # Import parameters
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(fname=args.path_params, tokens=tokens, device=device, dtype=dtype)
    L, q = params["bias"].shape
    
    # Select the sampler
    sampler = get_sampler(args.sampler)
    
    # Import data
    print(f"Loading data from {args.data}...")
    dataset = DatasetDCA(
        path_data=args.data,
        path_weights=args.weights,
        alphabet=tokens,
        clustering_th=args.clustering_seqid,
        no_reweighting=args.no_reweighting,
        device=device,
        dtype=dtype,
    )
    data = one_hot(dataset.data, num_classes=len(tokens)).to(dtype)
    weights = dataset.weights
    nmeasure = min(args.nmeasure, len(data))
    data_resampled = resample_sequences(data, weights, nmeasure)
    
    if args.pseudocount is None:
        args.pseudocount = 1. / weights.sum()
    print(f"Using pseudocount: {args.pseudocount}...")
    
    # Compute the mixing time
    print("Computing the mixing time...")
    results_mix = compute_mixing_time(
        sampler=sampler,
        data=data_resampled,
        params=params,
        n_max_sweeps=args.max_nsweeps,
        beta=args.beta,
    )
    mixing_time = results_mix["t_half"][-1]
    print(f"Measured mixing time (if converged): {mixing_time} sweeps")
    
    # Sample from random initialization
    print(f"Sampling for {args.nmix} * t_mix sweeps...")
    pbar = tqdm(
        total=args.nmix * mixing_time,
        colour="red",
        dynamic_ncols=True,
        leave=False,
        ascii="-#",
    )
    
    # Initialize the chains at random
    samples = init_chains(
        num_chains=args.ngen,
        L=L,
        q=q,
        device=device,
        dtype=dtype,
    )
    
    # Compute single and two-site frequencies of the data
    fi = get_freq_single_point(data=data, weights=weights, pseudo_count=args.pseudocount)
    fij = get_freq_two_points(data=data, weights=weights, pseudo_count=args.pseudocount)
    
    # Dictionary to store Pearson coefficient and slope along the sampling
    results_sampling = {
        "nsweeps" : [],
        "pearson" : [],
        "slope" : [],
    }
    
    # Sample for (args.nmix * mixing_time) sweeps starting from random initialization
    for i in range(args.nmix * mixing_time):
        pbar.update(1)
        samples = sampler(chains=samples, params=params, nsweeps=1, beta=args.beta)
        pi = get_freq_single_point(data=samples, weights=None, pseudo_count=0.)
        pij = get_freq_two_points(data=samples, weights=None, pseudo_count=0.)
        pearson, slope = get_correlation_two_points(fi=fi, pi=pi, fij=fij, pij=pij)
        results_sampling["nsweeps"].append(i)
        results_sampling["pearson"].append(pearson)
        results_sampling["slope"].append(slope)
    pbar.close()
    
    # Compute the energy of the samples
    print("Computing the energy of the samples...")
    energies = compute_energy(samples, params=params).cpu().numpy()
    
    print("Saving the samples...")
    headers = [f"sequence {i+1} | DCAenergy: {energies[i]:.3f}" for i in range(args.ngen)]
    write_fasta(
        fname=folder / Path(f"{args.label}_samples.fasta"),
        headers=headers,
        sequences=samples.argmax(-1).cpu().numpy(),
        numeric_input=True,
        remove_gaps=False,
        tokens=tokens,
    )
    
    print("Writing sampling log...")
    df_mix_log = pd.DataFrame.from_dict(results_mix)    
    df_mix_log.to_csv(
        folder / Path(f"{args.label}_mix.log"),
        index=False
    )
    df_samp_log = pd.DataFrame.from_dict(results_sampling)    
    df_samp_log.to_csv(
        folder / Path(f"{args.label}_sampling.log"),
        index=False
    )
    
    print(f"Done, results saved in {str(folder)}")
    
    
if __name__ == "__main__":
    main()


from typing import Any
from pathlib import Path

from torch.utils.data import Dataset
import torch

from adabmDCA.fasta import (
    get_tokens,
    import_from_fasta,
    compute_weights,
)


class DatasetDCA(Dataset):
    """Dataset class for handling multi-sequence alignments data."""
    def __init__(
        self,
        path_data: str | Path,
        path_weights: str | Path | None = None,
        alphabet: str = "protein",
        clustering_th: float = 0.8,
        no_reweighting: bool = False,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
        message: bool = True,
    ):
        """Initialize the dataset.

        Args:
            path_data (str | Path): Path to multi sequence alignment in fasta format.
            path_weights (str | Path | None, optional): Path to the file containing the importance weights of the sequences. If None, the weights are computed automatically.
            alphabet (str, optional): Selects the type of encoding of the sequences. Default choices are ("protein", "rna", "dna"). Defaults to "protein".
            clustering_th (float, optional): Sequence identity threshold for clustering. Defaults to 0.8.
            no_reweighting (bool, optional): If True, the weights are not computed. Defaults to False.
            device (torch.device, optional): Device to be used. Defaults to "cpu".
            dtype (torch.dtype, optional): Data type of the dataset. Defaults to torch.float32.
            message (bool, optional): Print the import message. Defaults to True.
        """
        path_data = Path(path_data)
        self.names = None
        self.data = None
        self.device = device
        self.dtype = dtype
        
        # Select the proper encoding
        self.tokens = get_tokens(alphabet)
        
        # Automatically detects if the file is in fasta format and imports the data
        with open(path_data, "r") as f:
            first_line = f.readline()
        if first_line.startswith(">"):
            self.names, self.data = import_from_fasta(path_data, tokens=self.tokens, filter_sequences=True)
            self.data = torch.tensor(self.data, device=device, dtype=torch.int32)
            # Check if data is empty
            if len(self.data) == 0:
                raise ValueError(f"The input dataset is empty. Check that the alphabet is correct. Current alphabet: {alphabet}")
        else:
            raise KeyError("The input dataset is not in fasta format")
        
        # Computes the weights to be assigned to the data
        if no_reweighting:
            self.weights = torch.ones(len(self.data), device=device, dtype=dtype)
        elif path_weights is None:
            if message:
                print("Automatically computing the sequence weights...")
            self.weights = compute_weights(data=self.data, th=clustering_th, device=device, dtype=dtype)
        else:
            with open(path_weights, "r") as f:
                weights = [float(line.strip()) for line in f]
            self.weights = torch.tensor(weights, device=device, dtype=dtype)
        
        if message:
            print(f"Multi-sequence alignment imported: M = {self.data.shape[0]}, L = {self.data.shape[1]}, q = {self.get_num_states()}, M_eff = {int(self.weights.sum())}.")


    def __len__(self):
        return len(self.data)


    def __getitem__(self, idx: int) -> Any:
        sample = self.data[idx]
        weight = self.weights[idx]
        return (sample, weight)
    
    
    def get_num_residues(self) -> int:
        """Returns the number of residues (L) in the multi-sequence alignment.

        Returns:
            int: Length of the MSA.
        """
        return self.data.shape[1]
    
    
    def get_num_states(self) -> int:
        """Returns the number of states (q) in the alphabet.

        Returns:
            int: Number of states.
        """
        return torch.max(self.data).item() + 1
    
    
    def get_effective_size(self) -> int:
        """Returns the effective size (Meff) of the dataset.

        Returns:
            int: Effective size of the dataset.
        """
        return int(self.weights.sum())
    
    
    def shuffle(self) -> None:
        """Shuffles the dataset.
        """
        perm = torch.randperm(len(self.data))
        self.data = self.data[perm]
        self.names = self.names[perm]
        self.weights = self.weights[perm]


import argparse
from pathlib import Path

import torch
import numpy as np
from tqdm import tqdm
import time

from adabmDCA.fasta import get_tokens
from adabmDCA.utils import init_chains, get_device, get_dtype, resample_sequences
from adabmDCA.io import load_params, load_chains, import_from_fasta
from adabmDCA.functional import one_hot
from adabmDCA.sampling import get_sampler
from adabmDCA.statmech import compute_energy
from adabmDCA.parser import add_args_tdint
from adabmDCA.resampling import compute_seqID 


# import command-line input arguments
def create_parser():
    parser = argparse.ArgumentParser(description="Computes the Entropy of the given model using a Thermodynamical Integration.")
    parser = add_args_tdint(parser)
    
    return parser


def main():
    
    # Parse arguments
    parser = create_parser()
    args = parser.parse_args()
    
    # Create output folder
    folder = Path(args.output)
    folder.mkdir(parents=True, exist_ok=True)

    print("\n" + "".join(["*"] * 10) + f" Computing model's entropy " + "".join(["*"] * 10) + "\n")
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)

    # Check if the data file exists
    if not Path(args.data).exists():
        raise FileNotFoundError(f"Data file {args.data} not found.")
    
    # Check if the parameters file exists
    if not Path(args.path_params).exists():
        raise FileNotFoundError(f"Parameters file {args.path_params} not found.")

    # Check if the target-sequence file exists
    if not Path(args.path_targetseq).exists():
        raise FileNotFoundError(f"Target Sequence file {args.path_targetseq} not found.")
    
    if args.label is not None:
        file_log = folder / Path(f"{args.label}.log")      
    else:
        file_log = folder / Path(f"td_integration.log")

    # select sampler
    sampler = get_sampler(args.sampler)

    # read and encode natural data
    tokens = get_tokens(args.alphabet)
    _, nat_data = import_from_fasta(args.data, tokens=tokens, filter_sequences=True)
    M, L, q = len(nat_data), len(nat_data[0]), len(tokens)
    nat_data = one_hot(torch.tensor(nat_data, device=device, dtype=torch.int32), num_classes=q).to(dtype)
    print(f"Number of sequences in the MSA: M={M}")
    print(f"Length of the MSA: L={L}")
    print(f"Number of Potts states: q={q}\n")

    # read parameters
    print(f"Loading parameters from {args.path_params}...")
    params = load_params(args.path_params, tokens=tokens, device=device, dtype=dtype)

    # read chains
    if args.path_chains is None:
        chains = init_chains(args.nchains, L, q, device=device, dtype=dtype)
    else:   
        chains = load_chains(args.path_chains, tokens=tokens)
        chains = one_hot(torch.tensor(chains, device=device, dtype=torch.int32), num_classes=q).to(dtype)
        if chains.shape[0] != args.nchains:
            chains = resample_sequences(chains, weights=torch.ones(chains.shape[0])/chains.shape[0], nextract=args.nchains)
    print(f"Number of chains set to {args.nchains}.")
        
    # target sequence
    _, targetseq = import_from_fasta(args.path_targetseq, tokens=tokens, filter_sequences=True)
    targetseq = one_hot(torch.tensor(targetseq, device=device, dtype=torch.int32), num_classes=q).to(dtype)
    if len(targetseq) != 1:
        print(f"Target sequence file contains more than one sequence. Using the first sequence as target sequence.")
        targetseq = targetseq[0]

    # initialize checkpoint
    template = "{0:<20} {1:<50}\n"  
    with open(file_log, "w") as f:
        if args.label is not None:
            f.write(template.format("label:", args.label))
        else:
            f.write(template.format("label:", "N/A"))
        
        f.write(template.format("input MSA:", str(args.data)))
        f.write(template.format("model path:", str(args.path_params)))
        f.write(template.format("target sequence:", str(args.path_targetseq)))
        f.write(template.format("alphabet:", args.alphabet))
        f.write(template.format("nchains:", args.nchains))
        f.write(template.format("nsweeps:", args.nsweeps))
        if args.nsweeps_theta is not None:
            f.write(template.format("nsweeps  theta:", str(args.nsweeps_theta)))
        if args.nsweeps_zero is not None:
            f.write(template.format("nsweeps zero:", str(args.nsweeps_zero)))
        f.write(template.format("nsteps:", args.nsteps))
        f.write(template.format("data type:", args.dtype))
        f.write(template.format("random seed:", args.seed))
        f.write("\n")
        # write the header of the log file
        logs = {
            "Epoch": 0,
            "Theta": 0.0,
            "Free Energy": 0.0,
            "Entropy": 0.0,
            "Time": 0.0
        }
        header_string = " ".join([f"{key:<15}" for key in logs.keys()])
        f.write(header_string + "\n")
    
    # Sampling to thermalize at theta = 0
    print("Thermalizing at theta = 0...")
    chains_0 = sampler(chains, params, args.nsweeps_zero) 
    ave_energy_0 = torch.mean(compute_energy(chains_0, params))

    # Sampling to thermalize at theta = theta_max
    print("Thermalizing at theta = theta_max...")
    theta_max = args.theta_max
    params_theta = {k : v.clone() for k, v in params.items()}
    params_theta["bias"] += theta_max * targetseq
    chains_theta = init_chains(args.nchains, L, q, device=device, dtype=dtype)
    chains_theta = sampler(chains_theta, params_theta, args.nsweeps_theta)
    seqID_max = compute_seqID(chains_theta, targetseq)
            
    # Find theta_max to generate 10% target sequences in the sample
    print("Finding theta_max to generate 10% target sequences in the sample...")
    p_wt =  (seqID_max == L).sum().item() / args.nchains # percentage of targetseq in the sample
    nsweep_find_theta = 100
    while p_wt <= 0.1:
        theta_max += 0.01 * theta_max
        print(f"{(p_wt * 100):.2f}% sequences collapse to WT")
        print(f"Number of sequences collapsed to WT is less than 10%. Increasing theta max to: {theta_max:.2f}...")
        params_theta["bias"] = params["bias"] + theta_max * targetseq
        chains_theta = sampler(chains_theta, params_theta, nsweep_find_theta)
        seqID = compute_seqID(chains_theta, targetseq)
        p_wt = (seqID == L).sum().item() / args.nchains
    
    # initiaize Thermodynamic Integration
    print("Starting Thermodynamic Integration...")
    int_step = args.nsteps
    nsweeps = args.nsweeps
    F_max = np.log(p_wt) + torch.mean(compute_energy(chains_theta[seqID == L], params_theta))
    thetas = torch.linspace(0, theta_max, int_step) 
    factor = theta_max / (2 * int_step)
    F, S, integral = F_max, 0, 0
    torch.set_printoptions(precision=2)

    # initialize progress bar
    pbar = tqdm(
        initial=max(0, thetas[0]),
        total=round(theta_max, 3),
        colour="red",
        dynamic_ncols=True,
        leave=False,
        ascii="-#",
        bar_format="{desc} {percentage:.2f}%[{bar}] Theta: {n:.3f}/{total_fmt} [{elapsed}]"
    )
    pbar.set_description(f"Step: {0} - SeqID: Nan - Entropy: Nan")

    time_start = time.time()

    for i, theta in enumerate(thetas):
        
        # sampling and compute seqID
        params_theta["bias"] = params["bias"] + theta * targetseq
        chains_theta = sampler(chains_theta, params_theta, nsweeps)
        seqID = compute_seqID(chains_theta, targetseq)
        mean_seqID = seqID.mean()
        
        # step of integration to compute entropy
        if i == 0 or i == int_step - 1:
            F += factor * torch.mean(seqID) 
            integral += factor * mean_seqID
        else:
            F += 2 * factor * mean_seqID
            integral += 2 * factor * mean_seqID
        S = ave_energy_0 - F
   
        # progress bar
        pbar.n = min(max(0, float(theta)), theta_max)
        pbar.set_description(f"Step: {i} - SeqID: {mean_seqID:.3f} - Entropy: {S:.2f}")

        # checkpoint
        logs["Epoch"] = i
        logs["Theta"] = float(theta)
        logs["Free Energy"] = F.item()
        logs["Entropy"] = S.item()
        logs["Time"] = time.time() - time_start
        with open(file_log, "a") as f:
            f.write(" ".join([f"{value:<15.3f}" if isinstance(value, float) else f"{value:<15}" for value in logs.values()]) + "\n")
        
    pbar.close()
    print(f"Process completed. Results saved in {file_log}.")


if __name__ == "__main__":
    main()


import torch


def _get_ortho(mat: torch.Tensor):
    """Orthonormalize the column vectors of a matrix.

    Args:
        mat (torch.Tensor): Matrix to orthonormalized. (a, b)

    Returns:
        torch.Tensor: Orthonormalized matrix. (a, b)
    """
    res = mat.clone()
    d = mat.shape[1]

    u0 = mat[:, 0] / mat[:, 0].norm()
    res[:, 0] = u0
    for i in range(1, d):
        ui = mat[:, i]
        for j in range(i):
            ui -= (ui @ res[:, j]) * res[:, j]
        res[:, i] = ui / ui.norm()
        
    return res


def _compute_U(
    M: torch.Tensor,
    num_directions: int,
    device: torch.device,
    dtype: torch.dtype,
) -> torch.Tensor:
    """Compute the first 'num_directions' right eigenvector of the dataset.

    Args:
        M (torch.Tensor): Dataset (num_samples, num_variables).
        num_directions (int): Number of principal components to compute.
        device (torch.device): Device to use.
        dtype (torch.dtype): Data type to use.
    
    Returns:
        torch.Tensor: Principal components (num_variables, num_directions).
    """
    num_samples, num_variables = M.shape
    max_iter = 100
    err_threshold = 1e-15
    curr_v = torch.rand(num_samples, num_directions, device=device, dtype=dtype) * 2 - 1
    u = torch.rand(num_variables, num_directions, device=device, dtype=dtype)
    curr_id_mat = torch.rand(num_directions, num_directions, device=device, dtype=dtype) * 2 - 1
    for n in range(max_iter):
        v = curr_v.clone()
        curr_v = M @ u
        if num_samples < num_variables:
            id_mat = (v.T @ curr_v) / num_samples
            curr_v = _get_ortho(curr_v)
        curr_u = M.T @ curr_v
        if num_variables <= num_samples:
            id_mat = (u.T @ curr_u) / num_samples
            curr_u = _get_ortho(curr_u)
        u = curr_u.clone()
        if (id_mat - curr_id_mat).norm() < err_threshold:
            break
        curr_id_mat = id_mat.clone()
    u = _get_ortho(u)

    return u

class Pca():
    def __init__(self):
        self.num_directions = 0
        self.U = None

    def fit(
        self,
        M: torch.Tensor,
        num_directions: int = 2,
    ) -> None:
        """Fit the PCA model to the data.

        Args:
            M (torch.Tensor): Data matrix (num_samples, num_variables).
            num_directions (int): Number of principal components to compute.
        """
        # Check that the input matrix is a 2D tensor
        if M.dim() != 2:
            raise ValueError("Input matrix must be a 2D tensor.")
        self.num_directions = num_directions
        self.U = _compute_U(M, self.num_directions, M.device, M.dtype)

    def transform(
        self,
        M: torch.Tensor,
    ) -> torch.Tensor:
        """Projects the data onto the principal components.

        Args:
            M (torch.Tensor): Data matrix (num_samples, num_variables).

        Returns:
            torch.Tensor: Projected data matrix (num_samples, num_directions).
        """
        return M @ self.U

    def fit_transform(
        self,
        M: torch.Tensor,
        num_directions: int = 2,
    ) -> torch.Tensor:
        """Fit the PCA model to the data and project the data onto the principal components.

        Args:
            M (torch.Tensor): Data matrix (num_samples, num_variables).
            num_directions (int): Number of principal components to compute.

        Returns:
            torch.Tensor: Projected data matrix (num_samples, num_directions).
        """
        self.fit(M, num_directions)
        return self.transform(M)


from typing import Dict

import torch

from adabmDCA.functional import one_hot
    
    
def set_zerosum_gauge(params: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    """Sets the zero-sum gauge on the coupling matrix.
    
    Args:
        params (Dict[str, torch.Tensor]): Parameters of the model.
        
    Returns:
        Dict[str, torch.Tensor]: Parameters with fixed gauge.
    """
    coupling_matrix = params["coupling_matrix"]
    coupling_matrix -= coupling_matrix.mean(dim=1, keepdim=True) + \
                       coupling_matrix.mean(dim=3, keepdim=True) - \
                       coupling_matrix.mean(dim=(1, 3), keepdim=True)
    
    params["coupling_matrix"] = coupling_matrix
    
    return params


def init_parameters(fi: torch.Tensor) -> Dict[str, torch.Tensor]:
    """Initialize the parameters of the DCA model.

    Args:
        fi (torch.Tensor): Single-point frequencies of the data.

    Returns:
        Dict[str, torch.Tensor]: Parameters of the model.
    """
    L, q = fi.shape
    params = {}
    params["bias"] = torch.log(fi)
    params["coupling_matrix"] = torch.zeros((L, q, L, q), device=fi.device, dtype=fi.dtype)
    
    return params


def init_chains(
    num_chains: int,
    L: int,
    q: int,
    device: torch.device,
    dtype: torch.dtype = torch.float32,
    fi: torch.Tensor | None = None,
) -> torch.Tensor:
    """Initialize the chains of the DCA model. If 'fi' is provided, the chains are sampled from the
    profile model, otherwise they are sampled uniformly at random.

    Args:
        num_chains (int): Number of parallel chains.
        L (int): Length of the MSA.
        q (int): Number of values that each residue can assume.
        device (torch.device): Device where to store the chains.
        dtype (torch.dtype, optional): Data type of the chains. Defaults to torch.float32.
        fi (torch.Tensor | None, optional): Single-point frequencies. Defaults to None.

    Returns:
        torch.Tensor: Initialized parallel chains in one-hot encoding format.
    """
    if fi is None:
        chains = torch.randint(low=0, high=q, size=(num_chains, L), device=device)
    else:
        chains = torch.multinomial(fi, num_samples=num_chains, replacement=True).to(device=device).T
    
    return one_hot(chains, num_classes=q).to(dtype)


def get_mask_save(L: int, q: int, device: torch.device) -> torch.Tensor:
    """Returns the mask to save the upper-triangular part of the coupling matrix.
    
    Args:
        L (int): Length of the MSA.
        q (int): Number of values that each residue can assume.
        device (torch.device): Device where to store the mask.
        
    Returns:
        torch.Tensor: Mask.
    """
    mask_save = torch.ones((L, q, L, q), dtype=torch.bool, device=device)
    idx1_rm, idx2_rm = torch.tril_indices(L, L, offset=0)
    mask_save[idx1_rm, :, idx2_rm, :] = 0
    
    return mask_save


@torch.jit.script
def systematic_resampling(
    chains: torch.Tensor,
    weights: torch.Tensor,
) -> torch.Tensor:
    """Performs the systematic resampling of the chains according to their relative weight.

    Args:
        chains (torch.Tensor): Chains.
        weights (torch.Tensor): Weights of the chains.

    Returns:
        torch.Tensor: Resampled chains.
    """
    num_chains = len(chains)
    device = chains.device
    # Normalize the weights
    weights = weights.view(-1) / weights.sum()
    weights_span = torch.cumsum(weights.double(), dim=0).float()
    rand_unif = torch.rand(size=(1,), device=device)
    arrow_span = (torch.arange(num_chains, device=device) + rand_unif) / num_chains
    mask = (weights_span.reshape(num_chains, 1) >= arrow_span).sum(1)
    counts = torch.diff(mask, prepend=torch.tensor([0], device=device))
    chains = torch.repeat_interleave(chains, counts, dim=0)

    return chains


def resample_sequences(
    data: torch.Tensor,
    weights: torch.Tensor,
    nextract: int,
) -> torch.Tensor:
    """Extracts nextract sequences from data with replacement according to the weights.
    
    Args:
        data (torch.Tensor): Data array.
        weights (torch.Tensor): Weights of the sequences.
        nextract (int): Number of sequences to be extracted.

    Returns:
        torch.Tensor: Extracted sequences.
    """
    weights = weights.view(-1)
    indices = torch.multinomial(weights, nextract, replacement=True)
    
    return data[indices]


def get_device(device: str, message: bool = True) -> torch.device:
    """Returns the device where to store the tensors.
    
    Args:
        device (str): Device to be used.
        message (bool, optional): Print the device. Defaults to True.
        
    Returns:
        torch.device: Device.
    """
    if "cuda" in device and torch.cuda.is_available():
        device = torch.device(device)
        if message:
            print(f"Running on {torch.cuda.get_device_name(device)}")
        return device
    else:
        if message:
            print("Running on CPU")
        return torch.device("cpu")
    
    
def get_dtype(dtype: str) -> torch.dtype:
    """Returns the data type of the tensors.
    
    Args:
        dtype (str): Data type.
        
    Returns:
        torch.dtype: Data type.
    """
    if dtype == "float32":
        return torch.float32
    elif dtype == "float64":
        return torch.float64
    elif dtype == "float16":
        return torch.float16
    else:
        raise ValueError(f"Data type {dtype} not supported.")


from pathlib import Path
import importlib
import argparse
import numpy as np

import torch

from adabmDCA.dataset import DatasetDCA
from adabmDCA.fasta import get_tokens, import_from_fasta
from adabmDCA.io import load_chains, load_params
from adabmDCA.stats import get_freq_single_point, get_freq_two_points
from adabmDCA.utils import init_chains, init_parameters, get_device, get_dtype
from adabmDCA.parser import add_args_train
from adabmDCA.sampling import get_sampler
from adabmDCA.functional import one_hot
from adabmDCA.checkpoint import get_checkpoint


# import command-line input arguments
def create_parser():
    # Important arguments
    parser = argparse.ArgumentParser(description='Train a DCA model.')
    parser = add_args_train(parser)
    
    return parser


def main():
    
    # Load parser, training dataset and DCA model
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Training {args.model} model " + "".join(["*"] * 10) + "\n")
    # Set the device
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    template = "{0:<30} {1:<50}"
    print(template.format("Input MSA:", str(args.data)))
    print(template.format("Output folder:", str(args.output)))
    print(template.format("Alphabet:", args.alphabet))
    print(template.format("Learning rate:", args.lr))
    print(template.format("Number of sweeps:", args.nsweeps))
    print(template.format("Sampler:", args.sampler))
    print(template.format("Target Pearson Cij:", args.target))
    if args.pseudocount is not None:
        print(template.format("Pseudocount:", args.pseudocount))
    print(template.format("Random seed:", args.seed))
    print(template.format("Data type:", args.dtype))
    print("\n")
    
    # Check if the data file exist
    if not Path(args.data).exists():
        raise FileNotFoundError(f"Data file {args.data} not found.")
    
    if args.test is not None:
        if not Path(args.test).exists():
            raise FileNotFoundError(f"Test file {args.test} not found.")
    
    # Create the folder where to save the model
    folder = Path(args.output)
    folder.mkdir(parents=True, exist_ok=True)
    
    if args.label is not None:
        file_paths = {
            "log" : folder / Path(f"{args.label}.log"),
            "params" : folder / Path(f"{args.label}_params.dat"),
            "chains" : folder / Path(f"{args.label}_chains.fasta")
        }
        
    else:
        file_paths = {
            "log" : folder / Path(f"adabmDCA.log"),
            "params" : folder / Path(f"params.dat"),
            "chains" : folder / Path(f"chains.fasta")
        }
    
    # Import dataset
    print("Importing dataset...")
    dataset = DatasetDCA(
        path_data=args.data,
        path_weights=args.weights,
        alphabet=args.alphabet,
        clustering_th=args.clustering_seqid,
        no_reweighting=args.no_reweighting,
        device=device,
        dtype=dtype,
    )
    
    # Import the test dataset if provided
    if args.test is not None:
        print("Importing test dataset...")
        test_dataset = DatasetDCA(
            path_data=args.test,
            path_weights=None,
            alphabet=args.alphabet,
            clustering_th=args.clustering_seqid,
            no_reweighting=args.no_reweighting,
            device=device,
            dtype=dtype,
        )
        test_oh = one_hot(test_dataset.data, num_classes=dataset.get_num_states()).to(dtype)
        pseudocount_test = 1. / test_dataset.get_effective_size()
        fi_test = get_freq_single_point(data=test_oh, weights=test_dataset.weights, pseudo_count=pseudocount_test)
        fij_test = get_freq_two_points(data=test_oh, weights=test_dataset.weights, pseudo_count=pseudocount_test)
    else:
        fi_test = None
        fij_test = None
    
    DCA_model = importlib.import_module(f"adabmDCA.models.{args.model}")
    tokens = get_tokens(args.alphabet)
    
    # Save the weights if not already provided
    if args.weights is None:
        if args.label is not None:
            path_weights = folder / f"{args.label}_weights.dat"
        else:
            path_weights = folder / "weights.dat"
        np.savetxt(path_weights, dataset.weights.cpu().numpy())
        print(f"Weights saved in {path_weights}")
        
    # Set the random seed
    torch.manual_seed(args.seed)
        
    # Shuffle the dataset
    dataset.shuffle()
    
    # Compute statistics of the data
    L = dataset.get_num_residues()
    q = dataset.get_num_states()
    
    if args.pseudocount is None:
        args.pseudocount = 1. / dataset.get_effective_size()
        print(f"Pseudocount automatically set to {args.pseudocount}.")
        
    data_oh = one_hot(dataset.data, num_classes=q).to(dtype)
    fi_target = get_freq_single_point(data=data_oh, weights=dataset.weights, pseudo_count=args.pseudocount)
    fij_target = get_freq_two_points(data=data_oh, weights=dataset.weights, pseudo_count=args.pseudocount) 
    
    # Initialize parameters and chains
    if args.path_params:
        print("Loading parameters...")
        tokens = get_tokens(args.alphabet)
        params = load_params(fname=args.path_params, tokens=tokens, device=device, dtype=dtype)
        mask = torch.zeros(size=(L, q, L, q), dtype=torch.bool, device=device)
        mask[torch.nonzero(params["coupling_matrix"])] = 1
        
    else:
        params = init_parameters(fi=fi_target)
        
        if args.model in ["bmDCA", "edDCA"]:
            mask = torch.ones(size=(L, q, L, q), dtype=torch.bool, device=device)
            mask[torch.arange(L), :, torch.arange(L), :] = 0
            
        else:
            mask = torch.zeros(size=(L, q, L, q), device=device, dtype=torch.bool)
    
    if args.path_chains:
        print("Loading chains...")
        chains, log_weights = load_chains(fname=args.path_chains, tokens=dataset.tokens, load_weights=True)
        chains = one_hot(
            torch.tensor(chains, device=device),
            num_classes=q,
        ).to(dtype)
        log_weights = torch.tensor(log_weights, device=device, dtype=dtype)
        args.nchains = chains.shape[0]
        print(f"Loaded {args.nchains} chains.")
        
    else:
        print(f"Number of chains set to {args.nchains}.")
        chains = init_chains(num_chains=args.nchains, L=L, q=q, fi=fi_target, device=device, dtype=dtype)
        log_weights = torch.zeros(size=(args.nchains,), device=device, dtype=dtype)
        
    # Select the sampling function
    sampler = get_sampler(args.sampler)
    
    print("\n")
    
    checkpoint = get_checkpoint(args.checkpoints)(
        file_paths=file_paths,
        tokens=tokens,
        args=args,
        params=params,
        chains=chains,
        max_epochs=args.nepochs,
        target_acc_rate=args.target_acc_rate,
        use_wandb=args.wandb,
    )

    DCA_model.fit(
        sampler=sampler,
        fij_target=fij_target,
        fi_target=fi_target,
        fi_test=fi_test,
        fij_test=fij_test,
        params=params,
        mask=mask,
        chains=chains,
        log_weights=log_weights,
        tokens=tokens,
        target_pearson=args.target,
        pseudo_count=args.pseudocount,
        nsweeps=args.nsweeps,
        nepochs=args.nepochs,
        lr=args.lr,
        factivate=args.factivate,
        gsteps=args.gsteps,
        drate=args.drate,
        target_density=args.density,
        checkpoint=checkpoint,
    )
    
    
if __name__ == "__main__":
    main()


import time
from tqdm.autonotebook import tqdm
from typing import Callable

import torch

from adabmDCA.stats import get_freq_single_point, get_freq_two_points, get_correlation_two_points
from adabmDCA.training import train_graph
from adabmDCA.utils import get_mask_save
from adabmDCA.graph import activate_graph, compute_density
from adabmDCA.statmech import compute_log_likelihood, compute_entropy, _compute_ess
from adabmDCA.checkpoint import Checkpoint


def fit(
    sampler: Callable,
    fi_target: torch.Tensor,
    fij_target: torch.Tensor,
    params: dict,
    mask: torch.Tensor,
    chains: torch.Tensor,
    log_weights: torch.Tensor,
    target_pearson: float,
    nsweeps: int,
    nepochs: int,
    pseudo_count: float,
    lr: float,
    factivate: float,
    gsteps: int,
    fi_test: torch.Tensor | None = None,
    fij_test: torch.Tensor | None = None,
    checkpoint: Checkpoint | None = None,
    *args, **kwargs,
) -> None:
    """
    Fits an eaDCA model on the training data and saves the results in a file.

    Args:
        sampler (Callable): Sampling function to be used.
        fi_target (torch.Tensor): Single-point frequencies of the data.
        fij_target (torch.Tensor): Two-point frequencies of the data.
        params (dict): Initialization of the model's parameters.
        mask (torch.Tensor): Initialization of the coupling matrix's mask.
        chains (torch.Tensor): Initialization of the Markov chains.
        log_weights (torch.Tensor): Log-weights of the chains. Used to estimate the log-likelihood.
        target_pearson (float): Pearson correlation coefficient on the two-points statistics to be reached.
        nsweeps (int): Number of Monte Carlo steps to update the state of the model.
        nepochs (int): Maximum number of epochs to be performed. Defaults to 50000.
        pseudo_count (float): Pseudo count for the single and two points statistics. Acts as a regularization.
        lr (float): Learning rate.
        factivate (float): Fraction of inactive couplings to activate at each step.
        gsteps (int): Number of gradient updates to be performed on a given graph.
        fi_test (torch.Tensor | None): Single-point frequencies of the test data. Defaults to None.
        fij_test (torch.Tensor | None): Two-point frequencies of the test data. Defaults to None.
        checkpoint (Checkpoint | None): Checkpoint class to be used to save the model. Defaults to None.
    """
    
    # Check the input sizes
    if fi_target.dim() != 2:
        raise ValueError("fi_target must be a 2D tensor")
    if fij_target.dim() != 4:
        raise ValueError("fij_target must be a 4D tensor")
    if chains.dim() != 3:
        raise ValueError("chains must be a 3D tensor")
    
    device = fi_target.device
    dtype = fi_target.dtype
    checkpoint.checkpt_interval = 10 # Save the model every 10 graph updates
    checkpoint.max_epochs = nepochs
    
    graph_upd = 0
    density = compute_density(mask) * 100
    L, q = fi_target.shape
        
    # Mask for saving only the upper-diagonal matrix
    mask_save = get_mask_save(L, q, device=device)
    
    # log_weights used for the online computing of the log-likelihood
    logZ = (torch.logsumexp(log_weights, dim=0) - torch.log(torch.tensor(len(chains), device=device, dtype=dtype))).item()
    
    # Compute the single-point and two-points frequencies of the simulated data
    pi = get_freq_single_point(data=chains)
    pij = get_freq_two_points(data=chains)
    pearson = max(0, float(get_correlation_two_points(fij=fij_target, pij=pij, fi=fi_target, pi=pi)[0]))
    
    # Number of active couplings
    nactive = mask.sum()
    
    # Training loop
    time_start = time.time()
    log_likelihood = compute_log_likelihood(fi=fi_target, fij=fij_target, params=params, logZ=logZ)
    entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
        
    pbar = tqdm(initial=max(0, float(pearson)), total=target_pearson, colour="red", dynamic_ncols=True, ascii="-#",
                bar_format="{desc}: {percentage:.2f}%[{bar}] Pearson: {n:.3f}/{total_fmt} [{elapsed}]")
    pbar.set_description(f"Graph updates: {graph_upd} - Density: {density:.3f}% - New active couplings: {0} - LL: {log_likelihood:.3f}")
    
    while pearson < target_pearson:
        
        # Old number of active couplings
        nactive_old = nactive
        
        # Compute the two-points frequencies of the simulated data with pseudo-count
        pij_Dkl = get_freq_two_points(data=chains, weights=None, pseudo_count=pseudo_count)
        
        # Update the graph
        nactivate = int(((L**2 * q**2) - mask.sum().item()) * factivate)
        mask = activate_graph(
            mask=mask,
            fij=fij_target,
            pij=pij_Dkl,
            nactivate=nactivate,
        )
        
        # New number of active couplings
        nactive = mask.sum()
        
        # Bring the model at convergence on the graph
        chains, params, log_weights = train_graph(
            sampler=sampler,
            chains=chains,
            mask=mask,
            fi=fi_target,
            fij=fij_target,
            params=params,
            nsweeps=nsweeps,
            lr=lr,
            max_epochs=gsteps,
            target_pearson=target_pearson,
            log_weights=log_weights,
            check_slope=False,
            checkpoint=None,
            progress_bar=False,
        )

        graph_upd += 1
        
        # Compute the single-point and two-points frequencies of the simulated data
        pi = get_freq_single_point(data=chains)
        pij = get_freq_two_points(data=chains)
        
        # Compute statistics of the training
        pearson, slope = get_correlation_two_points(fij=fij_target, pij=pij, fi=fi_target, pi=pi)
        density = compute_density(mask) * 100
        logZ = (torch.logsumexp(log_weights, dim=0) - torch.log(torch.tensor(len(chains), device=device, dtype=dtype))).item()
        log_likelihood = compute_log_likelihood(fi=fi_target, fij=fij_target, params=params, logZ=logZ)
        pbar.set_description(f"Graph updates: {graph_upd} - Density: {density:.3f}% - New active couplings: {int(nactive - nactive_old)} - LL: {log_likelihood:.3f}")

        # Save the model if a checkpoint is reached
        if checkpoint.check(graph_upd, params, chains):
            entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
            ess = _compute_ess(log_weights)
            checkpoint.log(
                {
                    "Epochs": graph_upd,
                    "Pearson": pearson,
                    "Slope": slope,
                    "LL_train": log_likelihood,
                    "ESS": ess,
                    "Entropy": entropy,
                    "Density": density,
                    "Time": time.time() - time_start,
                }
            )
            if fi_test is not None and fij_test is not None:
                log_likelihood_test = compute_log_likelihood(fi=fi_test, fij=fij_test, params=params, logZ=logZ)
                checkpoint.log({"LL_test": log_likelihood_test})
            else:
                checkpoint.log({"LL_test": float("nan")})
            
            checkpoint.save(
                params=params,
                mask=torch.logical_and(mask, mask_save),
                chains=chains,
                log_weights=log_weights,
                )
        pbar.n = min(max(0, float(pearson)), target_pearson)

    entropy = compute_entropy(chains=chains, params=params, logZ=logZ)
    ess = _compute_ess(log_weights)
    checkpoint.log(
        {
            "Epochs": graph_upd,
            "Pearson": pearson,
            "Slope": slope,
            "LL_train": log_likelihood,
            "ESS": ess,
            "Entropy": entropy,
            "Density": density,
            "Time": time.time() - time_start,
        }
    )
    if fi_test is not None and fij_test is not None:
        log_likelihood_test = compute_log_likelihood(fi=fi_test, fij=fij_test, params=params, logZ=logZ)
        checkpoint.log({"LL_test": log_likelihood_test})
    else:
        checkpoint.log({"LL_test": float("nan")})
                
    checkpoint.save(
        params=params,
        mask=torch.logical_and(mask, mask_save),
        chains=chains,
        log_weights=log_weights,
        )
    print(f"Completed, model parameters saved in {checkpoint.file_paths['params']}")
    pbar.close()